local SETTINGS = {
	colors = {
		real = {r=1,g=0,b=0},
		energy = {r=0,g=1,b=1},
		aura = {r=0,g=1,b=0},
		default = {r=1,g=1,b=0},
	},
	font = STANDARD_TEXT_FONT,
	size = 18,
}

local GetTime = GetTime

local function fTime(s)
	if s >= 86400 then return '%dd',s/86400 end
	if s >= 3600 then return '%dh',s/3600 end
	if s >= 60 then return '%dm',s/60 end
	if s >= 10 then return '%d',s end
	return '%.1f',s
end

-- Glint animation handler --
local aniCount = 5
local duration = 0.2
local animations = {}
for i=1, aniCount do
	local frame = CreateFrame('Frame')
	local tex = frame:CreateTexture()
	tex:SetTexture([[Interface\Cooldown\star4]])
	tex:SetAlpha(0)
	tex:SetAllPoints()
	tex:SetBlendMode('ADD')
	local group = tex:CreateAnimationGroup()
	local alpha = group:CreateAnimation('Alpha')
	alpha:SetOrder(1)
	alpha:SetChange(1)
	alpha:SetDuration(0)
	local scale = group:CreateAnimation('Scale')
	scale:SetOrder(1)
	scale:SetScale(1.5,1.5)
	scale:SetDuration(0)
	local scale2 = group:CreateAnimation('Scale')
	scale2:SetOrder(2)
	scale2:SetScale(0,0)
	scale2:SetDuration(duration)
	local spin = group:CreateAnimation('Rotation')
	spin:SetOrder(2)
	spin:SetDegrees(90)
	spin:SetDuration(duration)
	animations[i] = {frame = frame, group = group}
end

local aniNum = 1
local function animate(button)
	if not button:IsVisible() then return end
	local animation = animations[aniNum]
	local frame,group = animation.frame,animation.group
	group:Stop()
	frame:SetFrameStrata(button:GetFrameStrata())
	frame:SetFrameLevel(button:GetFrameLevel() + 10)
	frame:SetAllPoints(button)
	group:Play()
	aniNum = (aniNum % aniCount) + 1
end

-- Cooldown spinner and action timers --
local timers,actions = {},{}
local function CooldownTimer(parent)
	if timers[parent] then return timers[parent] end
	local timer = CreateFrame('frame', nil, parent)
	timer:SetAllPoints(parent)
	timer:SetFrameLevel(parent:GetFrameLevel() + 3)
	
	timer.text = timer:CreateFontString(nil, 'OVERLAY')
	--timer.text:SetShadowOffset(1,-1)
	timer.text:SetFont(SETTINGS.font or STANDARD_TEXT_FONT, SETTINGS.size or 18, 'OUTLINE')
	
	timer.start = 0
	timer.duration = 0
	
	timers[parent] = timer
	return timer
end

local function ActionTimer(parent)
	if actions[parent] then return actions[parent] end
	local timer = CreateFrame('frame', nil, parent)
	timer:SetAllPoints(parent)
	timer:SetFrameLevel(parent:GetFrameLevel() + 3)
	
	timer.text = timer:CreateFontString(nil, 'OVERLAY')
	--timer.text:SetShadowOffset(1,-1)
	timer.text:SetFont(SETTINGS.font or STANDARD_TEXT_FONT, SETTINGS.size or 18, 'OUTLINE')
	timer.text:SetPoint('CENTER')
	
	timer.border = timer:CreateTexture(nil, 'BACKGROUND')
	timer.border:SetTexture([[Interface\Buttons\CheckButtonHilight]])
	timer.border:SetAllPoints()
	timer.border:SetBlendMode('add')
	--t:SetVertexColor(0, 0.25, 1, 1)
	
	timer.action = parent:GetAttribute('action') or ActionButton_CalculateAction(parent)
	
	actions[parent] = timer
	return timer
end

local function SetTimer(parent, start, duration, action)
	if not parent then return end
	local timer = CooldownTimer(parent)
	timer.start = start or 0
	timer.duration = duration or 0
end

-- Fires normally when a cooldown spinner is set, except on action buttons, which behave spuriously
hooksecurefunc(getmetatable(ActionButton1.cooldown).__index, 'SetCooldown', function(cooldown, start, duration, enable)
	if cooldown:GetParent():GetAttribute('type') == 'action' then return end
	SetTimer(cooldown, start, duration, 'spinner')
end)

-- Hook for action buttons created by an addon like bartender
hooksecurefunc(getmetatable(ActionButton1).__index, 'SetAttribute', function(button, attribute, slot)
	if attribute ~= 'action' or not HasAction(slot) then return end
	if not actions[button] then
		ActionTimer(button)
	end
	actions[button].action = slot
end)

-- Loop through every frame in the interface to register existing action buttons
local function GetActionButtons(...)
	for i = 1, select('#',...) do
		local frame = select(i, ...)
		if frame:GetNumChildren() > 0 then
			GetActionButtons(frame:GetChildren())
		end
		if frame:GetAttribute('type') == 'action' then
			ActionTimer(frame)
		end
	end
end
--GetActionButtons(UIParent:GetChildren())

-- Play Glint animation on any action buttons that match
hooksecurefunc('UseAction', function(slot1, target, button)
	for button,timer in pairs(actions) do
		--local slot2 = button:GetAttribute('action') or ActionButton_CalculateAction(button)
		if slot1 == timer.action and button:IsVisible() then
			animate(button)
		end
	end
end)


local function SetTimerTime(timer,left,color)
	local width = timer:GetWidth() or 0
	if not timer:IsVisible() or width <= 0 or left <= 0 then return end
	if timer.border and color then
		if IsUsableAction(timer.action) and color == SETTINGS.colors.aura then
			timer.border:SetVertexColor(SETTINGS.colors.aura.r,SETTINGS.colors.aura.g,SETTINGS.colors.aura.b,1)
		else
			timer.border:SetVertexColor(SETTINGS.colors.real.r,SETTINGS.colors.real.g,SETTINGS.colors.real.b,1)
		end
	end
	if not color then color = SETTINGS.colors.default end
	timer.text:SetTextColor(color.r or 1, color.g or 1, color.b or 0, 1)
	timer.text:ClearAllPoints()
	-- todo: only resize font strings when the frame resizes instead of setting the font every single update
	-- :HookScript('OnSizeChanged', fixFonts)
	if width < 26 then
		if left < 60 then
			timer.text:SetPoint("CENTER", timer, "TOPRIGHT")
			timer.text:SetFont(SETTINGS.font or STANDARD_TEXT_FONT, width*(SETTINGS.size/26), "OUTLINE")
			timer.text:SetFormattedText("%d", left)
		else
			timer.text:SetText()
		end
	else
		timer.text:SetPoint('CENTER')
		if left < 600 then
			timer.text:SetFont(SETTINGS.font or STANDARD_TEXT_FONT, width*(SETTINGS.size/36), "OUTLINE")
		else
			timer.text:SetFont(SETTINGS.font or STANDARD_TEXT_FONT, width*(SETTINGS.size/45), "OUTLINE")
		end
		timer.text:SetFormattedText(fTime(left))
	end
end


--\\---------------------------\\--
--//   MANY AURAS, HANDLE IT   //--
--\\---------------------------\\--
local GetSpellInfo = GetSpellInfo
local auras = {} -- [aura name] = expiration time
local SharedAuras = { -- [aura name] = psuedo group
	[1130] = 1130, -- Hunter's Mark (Hunter)
}
local StickyAuras = {} -- things to keep track of regardless of target, like my crowd controls, single-target abilities
local function updateAuras()
	wipe(auras) -- lazy
	local unit = (UnitExists("target") and (UnitIsPlayer("target") or UnitCanAttack("player", "target"))) and "target" or "player"
	if UnitIsFriend("player", unit) then
		for i = 1,40 do
			local name, _, _, _, _, _, expires, caster = UnitBuff(unit, i)
			--local name, rank, icon, count, dispelType, duration, expires, caster, isStealable, shouldConsolidate, spellID, canApplyAura, isBossDebuff, value1, value2, value3 = UnitBuff(unit, i)
			if not name then break end
			if caster == "player" and expires > 0 then
				if (auras[name] and expires > auras[name]) or not auras[name] then
					auras[name] = expires
				end
			elseif SharedAuras[name] and expires > 0 then
				auras[SharedAuras[name]] = expires
			end
		end
		for i = 1,40 do
			local name, _, _, _, _, _, expires, caster, _, _, spellID = UnitBuff("player", i)
			if not name then break end
			if caster == "player" and expires > 0 and SpellIsSelfBuff(spellID) then
				if (auras[name] and expires > auras[name]) or not auras[name] then
					auras[name] = expires
				end
			end
		end
	else
		for i = 1,40 do
			local name, _, _, _, _, _, expires, caster = UnitDebuff(unit, i)
			--local name, rank, icon, count, dispelType, duration, expires, caster, isStealable, shouldConsolidate, spellID, canApplyAura, isBossDebuff, value1, value2, value3 = UnitDebuff(unit, i)
			if not name then break end
			if caster == "player" and expires > 0 then
				if (auras[name] and expires > auras[name]) or not auras[name] then
					auras[name] = expires
				end
			elseif SharedAuras[name] and expires > 0 then
				auras[SharedAuras[name]] = expires
			end
		end
		for i = 1,40 do
			local name, _, _, _, _, _, expires, caster = UnitBuff("player", i)
			--local name, rank, icon, count, dispelType, duration, expires, caster, isStealable, shouldConsolidate, spellID, canApplyAura, isBossDebuff, value1, value2, value3 = UnitBuff("player", i)
			if not name then break end
			if caster == "player" and expires > 0 then
				if (auras[name] and expires > auras[name]) or not auras[name] then
					auras[name] = expires
				end
			elseif SharedAuras[name] and expires > 0 then
				auras[SharedAuras[name]] = expires
			end
		end
	end
end

local _,CLASS = UnitClass('player')
local function GetTotemCooldown(totem)
	for i=1,4 do
		local haveTotem, name, startTime, duration = GetTotemInfo(i)
		if haveTotem and name == totem then
			return (startTime + duration) or 0
		end
	end
	return 0
end

local f,ts = CreateFrame('frame'),0
f:SetScript('OnUpdate', function(self, e)
	ts = ts + e
	if ts >= 0.1 then
		local now,left,parent,timer = GetTime(),0
		for parent,timer in pairs(timers) do
			if timer.duration > 2.2 then
				left = timer.start + timer.duration - now
				SetTimerTime(timer,left)
			end
		end
		for button,timer in pairs(actions) do
			local slot = timer.action or button:GetAttribute('action') or ActionButton_CalculateAction(button)
			if HasAction(slot) and button:IsVisible() then
				local start, duration = GetActionCooldown(slot)
				
				local realCooldown, energyCooldown, auraDuration = 0, 0, 0
				local btype, id = GetActionInfo(slot)
				if btype == 'macro' then _,_,id = GetMacroSpell(id) end
				if btype == 'item' then
					local name = GetItemInfo(id)
					if name and auras[name] then
						auraDuration = auras[name]-now
					end
				elseif btype == 'spell' or (btype == 'macro' and id) then
					local name, _, _, cost, _, power = GetSpellInfo(id)
					if name then
						if auras[name] then
							auraDuration = auras[name] - now
							--SpellIsSelfBuff(id)
						elseif SharedAuras[name] and auras[SharedAuras[name]] then
							auraDuration = auras[SharedAuras[name]] - now
						end
						
						------------------------- Totem Timer Test - MultiCastActionButton doesn't update until a slot change, this could be a sign of a bigger issue
						if CLASS == 'SHAMAN' then
							local totemTime = GetTotemCooldown(name)
							if totemTime > 0 then
								auraDuration = totemTime - now
							end
						end
						------------------------ Totem Timer Test
						
						if UnitPowerType("player") == power and cost > 0 then -- because GetPowerRegen only works for the active energy type
							local currentPower = UnitPower("player", power)
							if cost > currentPower then
								local passiveRegen, activeRegen = GetPowerRegen()
								local regenRate = UnitAffectingCombat("player") and activeRegen or passiveRegen
								local powerDifference = cost - currentPower
								if regenRate >= 1 then
									energyCooldown = powerDifference/regenRate
								end
							end
						end
					end
				end
				
				left = start + duration - now
				if left > energyCooldown and left > 0 and duration > 2.2 then
					SetTimerTime(timer,left,SETTINGS.colors.real)
				elseif energyCooldown > 0 then
					SetTimerTime(timer,energyCooldown,SETTINGS.colors.energy)
				elseif auraDuration > 0 then
					SetTimerTime(timer,auraDuration,SETTINGS.colors.aura)
				else
					timer.text:SetText()
					timer.border:SetVertexColor(1,1,1,0)
				end
			else
				timer.text:SetText()
				timer.border:SetVertexColor(1,1,1,0)
			end
		end
		ts = 0
	end
end)

local function recalculateActions()
	for button,timer in pairs(actions) do
		timer.action = button:GetAttribute('action') or ActionButton_CalculateAction(button)
	end
end

f:SetScript('OnEvent', function(self, event, ...)
	if event == 'UNIT_AURA' or event == 'PLAYER_TARGET_CHANGED' then
		updateAuras()
	elseif event == 'ACTIONBAR_SLOT_CHANGED' then
		recalculateActions()
	elseif event == 'PLAYER_LOGIN' then
		GetActionButtons(UIParent:GetChildren())
	end
end)
updateAuras()
f:RegisterEvent('UNIT_AURA')
f:RegisterEvent('PLAYER_TARGET_CHANGED')
f:RegisterEvent('ACTIONBAR_SLOT_CHANGED') -- looping through every action button on this event is overkill
f:RegisterEvent("PLAYER_LOGIN")